/*****************************************************************************
* wanhdlc.c	WAN HDLC Configuration Utility.
*
* Author:	Alex Feldman	<al.feldman@sangoma.com>
*
* Copyright:	(c) 2004 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ----------------------------------------------------------------------------
* Feb 19, 2004	Alex Feldman	Port to FreeBSD (Version 1.1)
* Jan 27, 2004	Alex Feldman	Initial version based on WANPIPE configurator.
*				(Version 1.0)
*****************************************************************************/

/*****************************************************************************
* Usage:
*****************************************************************************/

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <netinet/in.h>
#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <net/if.h>
# include <net/wanpipe.h>
# include <net/wanpipe_generic.h>
#else
# include <linux/sockios.h>
# include <linux/if.h>
# include <linux/hdlc.h>
# include <linux/if_wanpipe.h>
# include <linux/wanpipe.h>
# include <linux/wanpipe_generic.h>
#endif
#include "wanhdlcy.h"
#include "wanhdlc.h"

#define WANHDLC_VERSION	"1.1"
//#define DEBUG_MSG

extern int 	yy_reset();
extern int 	yyparse();
//extern int 	yyrestart(FILE*);
extern int	wanhdlc_help(void);
//extern FILE*	yyin;

char**	targv;
char**	arglim;

static int 		sock;
static struct ifreq	req;
struct if_settings 	ifsettings;

static sdla_te_cfg_t	te_cfg;
static wan_dev_conf_t	devconf;
#if defined(CONFIG_GENERIC_CISCO)
static cisco_proto	cisco;
#endif
#if defined(CONFIG_GENERIC_FR)
static fr_proto 	fr;
static fr_proto_pvc 	fr_pvc;
#endif
static wanlite_def_t	dev_def;

int wanhdlc_ifname(char* ifname);
int wanhdlc_channel_mask(unsigned long first_ts, unsigned long last_ts);
int wanhdlc_te_config(int arg);
int wanhdlc_set_te_config(int arg, unsigned long value, char* str);
int wanhdlc_dev_config(int arg, int value);
int wanhdlc_proto_cisco(int arg);
int wanhdlc_set_proto_cisco(int arg, unsigned long value);
int wanhdlc_proto_fr(int arg);
int wanhdlc_set_proto_fr(int arg, unsigned long value);
int wanhdlc_fr_pvc(int arg);
int wanhdlc_set_fr_pvc(int arg, unsigned long value, char* name);
int wanhdlc_proto_ppp(int arg);
char* wanhdlc_version(void);
int print_dbg_msg(const char *format, ...);

static int wanhdlc_print_config(char* ifname, wanlite_def_t* dev_def);

int print_dbg_msg(const char *format, ...)
{
#ifdef DEBUG_MSG
	va_list args;

	va_start(args, format);
	printf("DBG> ");
	vprintf(format, args);
	va_end(args);
#endif
	return 0;
}

char* wanhdlc_version(void)
{
	return WANHDLC_VERSION;	
}

int main(int argc, char* argv[])
{
	int ret;

	if (argc < 2){
		wanhdlc_help();
		return 1;
	}
	memset(&req, 0, sizeof(struct ifreq));

	yy_reset();

	targv = argv+1;
	arglim = argv+argc;

	if (!(ret = yyparse())){
		int	cmd;

		req.ifr_data = (void*)&ifsettings;
		if (ifsettings.type){
			if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0){
				printf("ERROR: Unable to create socket: %s\n",
							 strerror(errno));
				return errno;
		  	}
			cmd = SIOC_WANPIPE_DEVICE;
		}else{
			if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
				printf("ERROR: Unable to create socket: %s\n",
						 strerror(errno));
				return errno;
			}
			cmd = SIOC_WANPIPE_HWPROBE;
			req.ifr_data = (void*)&dev_def;
		}

		if ((ret = ioctl(sock, cmd, &req)) < 0){
			if (cmd == SIOC_WANPIPE_DEVICE){
				switch(ifsettings.type){
				case IF_PROTO_CISCO:
					printf("ERROR: Unable to set Cisco HDLC protocol information: %s\n",
							strerror(errno));
					break;
				case IF_PROTO_PPP:
					printf("ERROR: Unable to set PPP protocol information: %s\n",
							strerror(errno));
					break;
				case IF_IFACE_T1:
				case IF_IFACE_E1:
					printf("ERROR: Unable to set T1/E1 information: %s\n",
						      strerror(errno));
					break;
				case IF_IFACE_SYNC_SERIAL:
					printf("ERROR: Unable to set interface information: %s\n",
						      strerror(errno));
					break;
				case IF_PROTO_FR:
					printf("ERROR: Unable to set FR protocol information: %s\n",
						      strerror(errno));
					break;
				case IF_PROTO_FR_ADD_PVC:
					printf("ERROR: Unable to ADD PVC: %s\n", 
							strerror(errno));
					break;
				case IF_PROTO_FR_DEL_PVC:
					printf("ERROR: Unable to DEL PVC: %s\n", 
							strerror(errno));
					break;
				default:
					printf("ERROR: Unknown type (type=%04X)\n", 
						ifsettings.type);
					break;
				}
			}else if (cmd == SIOC_WANPIPE_HWPROBE){
				printf("ERROR: Failed to read hardware configuration: %s\n", 
						strerror(errno));
			}else{
				printf("ERROR: Unknown ioctl command (cmd=%08X)!\n",
						cmd);
			}
		}else if (cmd == SIOC_WANPIPE_HWPROBE){
			/* Print hardware configuration */
			wanhdlc_print_config(req.ifr_name, &dev_def);
		}
	}

	return ret;
}

int wanhdlc_ifname(char* ifname)
{
	print_dbg_msg("Device name %s\n", ifname);
	memset (&req, 0, sizeof(struct ifreq));
	strcpy(req.ifr_name, ifname); 
	return 0;
}

int wanhdlc_channel_mask(unsigned long first_ts, unsigned long last_ts)
{
	unsigned long	ts_mask = 0;
	int		i = 0;
	
	if (!first_ts){
		first_ts++;
	}
	for(i = first_ts; i <=last_ts; i++){
		if (i <= 0){
			printf("ERROR: Invalid timeslot %d!\n", i);
			return 0;
		}
		ts_mask |= (1 << (i-1)); 
	}
	return ts_mask;


}
int wanhdlc_te_config(int arg)
{
	ifsettings.ifs_te1 = (te1_settings*)&te_cfg;
	ifsettings.ifs_size = sizeof(sdla_te_cfg_t);
	if (!te_cfg.media) te_cfg.media = WANOPT_MEDIA_T1;
	if (te_cfg.media == WANOPT_MEDIA_T1){
		ifsettings.type = IF_IFACE_T1;
	}else{
		ifsettings.type = IF_IFACE_E1;
	}
	if (!te_cfg.lcode){
	       te_cfg.lcode = (te_cfg.media == WANOPT_MEDIA_T1) ? 
		       			WANOPT_LC_B8ZS : WANOPT_LC_HDB3;
	}
	if (!te_cfg.frame){
	       te_cfg.frame = (te_cfg.media == WANOPT_MEDIA_T1) ? 
		       			WANOPT_FR_ESF : WANOPT_FR_NCRC4;
	}
	if (!te_cfg.active_ch) te_cfg.active_ch = ENABLE_ALL_CHANNELS;
	if (!te_cfg.te_clock) te_cfg.te_clock = WANOPT_NORMAL_CLK;
	if (!te_cfg.lbo){
		te_cfg.lbo = (te_cfg.media == WANOPT_MEDIA_T1) ?
				WANOPT_T1_LBO_0_DB : WANOPT_E1_120;
	}
	
	print_dbg_msg("MEDIA %d\n", te_cfg.media);
	print_dbg_msg("LCODE %d\n", te_cfg.lcode);
	print_dbg_msg("FRAME %d\n", te_cfg.frame);
	print_dbg_msg("ACTIVE_CH %lX\n", te_cfg.active_ch);
	print_dbg_msg("TECLOCK %d\n", te_cfg.te_clock);
	print_dbg_msg("LBO %d\n", te_cfg.lbo);
	return 0;
}

int wanhdlc_set_te_config(int arg, unsigned long value, char* str)
{
	switch(arg){
	case HDLC_MEDIA:
		if (value == WANOPT_MEDIA_T1){
			ifsettings.type	= IF_IFACE_T1;
		}else{
			ifsettings.type	= IF_IFACE_E1;
		}
		te_cfg.media = (unsigned char)value;
		break;
		
	case HDLC_LCODE:
		te_cfg.lcode = (unsigned char)value;
		break;
		
	case HDLC_FRAME:
		te_cfg.frame = (unsigned char)value;
		break;
		
	case HDLC_CHANNEL:
		te_cfg.active_ch = (unsigned long)value;
		break;
		
	case HDLC_TECLOCK:
		te_cfg.te_clock = (unsigned char)value;
		break;

	case HDLC_LBO:
		te_cfg.lbo = (unsigned char)value;
		break;
	}
	return 0;
}

int wanhdlc_dev_config(int arg, int value)
{
	ifsettings.type	= IF_IFACE_SYNC_SERIAL;
	ifsettings.ifs_sync = (sync_serial_settings*)&devconf;
	ifsettings.ifs_size = sizeof(wan_dev_conf_t);
	switch(arg){
	case HDLC_INTERFACE:
		devconf.mask |= WANHDLC_CONF_INTERFACE;
		devconf.interface = value;
	  	print_dbg_msg("Interface %d\n", value);
		break;
	case HDLC_CLOCKING:
		devconf.mask |= WANHDLC_CONF_CLOCKING;
		devconf.bps = value;
	  	print_dbg_msg("Baud rate %d\n", value);
		break;
	case HDLC_MTU:
		devconf.mask |= WANHDLC_CONF_MTU;
		devconf.mtu = value;
	  	print_dbg_msg("MTU %d\n", value);
		break;
	case HDLC_UDPPORT:
		devconf.mask |= WANHDLC_CONF_UDPPORT;
		devconf.udp_port = value;
	  	print_dbg_msg("UDPPORT %d\n", value);
		break;
	case HDLC_TTL:
		devconf.mask |= WANHDLC_CONF_TTL;
		devconf.ttl = value;
	  	print_dbg_msg("TTL %d\n", value);
		break;
	}
	return 0;

}

int wanhdlc_proto_cisco(int arg)
{
#if defined(CONFIG_GENERIC_CISCO)
# if defined(__LINUX__)
	if (!cisco.interval) cisco.interval = 10;
	if (!cisco.timeout)  cisco.timeout = 25;
	print_dbg_msg("Interval %d\n", cisco.interval);
	print_dbg_msg("Timeout %d\n", cisco.timeout);
# endif

	ifsettings.type = IF_PROTO_CISCO;
	ifsettings.ifs_cisco = (void*)&cisco;
	ifsettings.ifs_size = sizeof(cisco_proto);
	return 0;
#else
	printf("ERROR: CISCO invalid protocol type!\n");
	return -EINVAL;
#endif
}

int wanhdlc_set_proto_cisco(int arg, unsigned long value)
{

#if defined(CONFIG_GENERIC_CISCO)
	switch(arg){
#if defined(__LINUX__)
	case HDLC_CISCO_INTERVAL:
		cisco.interval = value;
		break;
	case HDLC_CISCO_TIMEOUT:
		cisco.timeout = value;
		break;
#endif
	default:
		print_dbg_msg("ERROR: Invalid Cisco configuration parameter (%d)\n",
				arg);
		return -EINVAL;
	}
	return 0;
#else
	return -EINVAL;
#endif
}

int wanhdlc_proto_fr(int arg)
{
#if defined(CONFIG_GENERIC_FR)
	ifsettings.type = IF_PROTO_FR;
	ifsettings.ifs_fr = (void*)&fr;
	ifsettings.ifs_size = sizeof(fr_proto);

	if (!fr.t391) fr.t391 = 10; /* polling verification timer*/
	/* link integrity verification polling timer */
	if (!fr.t392) fr.t392 = 15;
	if (!fr.n391) fr.n391 = 6; /* full status polling counter*/
	if (!fr.n392) fr.n392 = 3; /* error threshold */
	if (!fr.n393) fr.n393 = 4; /* monitored events count */
	if (!fr.lmi)  fr.lmi = LMI_ANSI;

	print_dbg_msg("LMI %d\n", fr.lmi);
	print_dbg_msg("T391 %d\n", fr.t391);
	print_dbg_msg("T392 %d\n", fr.t392);
	print_dbg_msg("N391 %d\n", fr.n391);
	print_dbg_msg("N392 %d\n", fr.n392);
	print_dbg_msg("N393 %d\n", fr.n393);
	print_dbg_msg("DCE %d\n", fr.dce);
	return 0;
#else
	printf("ERROR: FR invalid protocol type!\n");
	return -EINVAL;
#endif
}

int wanhdlc_set_proto_fr(int arg, unsigned long value)
{
#if defined(CONFIG_GENERIC_FR)
	switch(arg){
	case HDLC_FR_LMI:
		fr.lmi = value;
		break;
	case HDLC_FR_T391:
		fr.t391 = value;
		break;
	case HDLC_FR_T392:
		fr.t392 = value;
		break;
	case HDLC_FR_N391:
		fr.n391 = value;
		break;
	case HDLC_FR_N392:
		fr.n392 = value;
		break;
	case HDLC_FR_N393:
		fr.n393 = value;
		break;
	case HDLC_FR_DCE:
		fr.dce = 1;
	default:
		printf("ERROR: Invalid FR protocol information (%d)\n",
				arg);
		return -EINVAL;
	}
	return 0;
#else
	return -EINVAL;
#endif
}

int wanhdlc_fr_pvc(int arg)
{
#if defined(CONFIG_GENERIC_FR)
	ifsettings.ifs_fr_pvc = &fr_pvc;
	ifsettings.ifs_size = sizeof(fr_proto_pvc);
# if !defined(KERNEL_V26)
	print_dbg_msg("Name %s\n", 
				fr_pvc.dev_name);
# endif
	print_dbg_msg("Action %s\n", 
			(ifsettings.type == IF_PROTO_FR_ADD_PVC) ? 
					"Create" : "Delete");
	print_dbg_msg("DLCI %d\n", 
			fr_pvc.dlci);
# if !defined(KERNEL_V26)
	print_dbg_msg("Inverse Arp %d\n", 
			fr_pvc.inverse_arp);
	print_dbg_msg("CIR BC %d\n", 
			fr_pvc.bc);
	print_dbg_msg("CIR Interval %d\n", 
			fr_pvc.interval);
	print_dbg_msg("CIR NFMark %d\n", 
			fr_pvc.nfmark);
# endif
	return 0;
#else
	printf("ERROR: FR invalid protocol type!\n");
	return -EINVAL;
#endif
}

int wanhdlc_set_fr_pvc(int arg, unsigned long value, char* name)
{
#if defined(CONFIG_GENERIC_FR)
	switch(arg){
# if !defined(KERNEL_V26)
	case HDLC_FR_PVC_NAME:
		strncpy(fr_pvc.dev_name, name, IFNAMSIZ);
		break;
# endif
	case HDLC_FR_PVC_ADD:
		ifsettings.type = IF_PROTO_FR_ADD_PVC;
		break;
	case HDLC_FR_PVC_DEL:
		ifsettings.type = IF_PROTO_FR_DEL_PVC;
		break;
	case HDLC_FR_PVC_DLCI:
		if (value > 1023){
			printf("ERROR: Invalid DCLI number %ld!\n",
					value);
			return -EINVAL;
		}
		fr_pvc.dlci = value;	
		break;
# if !defined(KERNEL_V26)
	case HDLC_FR_PVC_INVERSE_ARP:
		fr_pvc.inverse_arp = 1;
		break;
# endif
	case HDLC_FR_PVC_CIR:
		break;
# if !defined(KERNEL_V26)
	case HDLC_FR_PVC_BC:
		if (value > 5000000){
			printf("ERROR: Invalid BC number %ld!\n",
					value);
			return -EINVAL;
		}
		fr_pvc.bc = value;	
		break;
# endif
# if !defined(KERNEL_V26)
	case HDLC_FR_PVC_INTERVAL:
		if (value > 99){
			printf("ERROR: Invalid INTERVAL number %ld!\n",
					value);
			return -EINVAL;
		}
		fr_pvc.interval = value;	
		break;
# endif
# if !defined(KERNEL_V26)
	case HDLC_FR_PVC_NFMARK:
		if (value > 65535){
			printf("ERROR: Invalid NFMARK number %ld!\n",
					value);
			return -EINVAL;
		}
		fr_pvc.nfmark = value;	
		break;
# endif
	default:
		printf("ERROR: Invalid FR protocol information (%d)\n",
				arg);
		return -EINVAL;
	}
	return 0;
#else
	return -EINVAL;
#endif
}

int wanhdlc_proto_ppp(int arg)
{
#if defined(CONFIG_GENERIC_PPP)
	ifsettings.type = IF_PROTO_PPP;
	ifsettings.ifs_size = 0;
	return 0;
#else
	printf("ERROR: PPP invalid protocol type!\n");
	return -EINVAL;
#endif
}

static int wanhdlc_print_config(char* ifname, wanlite_def_t* dev_def)
{
       union {
                fr_proto	fr;
                cisco_proto	cisco;
        } proto;

	printf("\nHardware configuration for %s:\n", 
			ifname);
	printf("%s\n", dev_def->hwprobe);

#if defined(__LINUX__)
	req.ifr_settings.type = IF_GET_PROTO;
#endif
	ifsettings.type = IF_GET_PROTO;
	ifsettings.ifs_cisco = (void*)&proto;
	ifsettings.ifs_size = sizeof(proto);
	req.ifr_data = (void*)&ifsettings;
	if (ioctl(sock, SIOC_WANPIPE_DEVICE, &req) < 0){
		return -EINVAL;
	}

	printf("\nInterface configuration for %s:\n", 
			ifname);

	printf("DEVICE=%s\n", ifname);
	printf("BOARD=wanpipe\n");
	if (ifsettings.type != IF_PROTO_FR){
		printf("IPADDR=\n");
		printf("NETMASK=\n");
		printf("NETWORK=\n");
		printf("BROADCAST=\n");
		printf("POINTOPOINT=\n");
	}
	printf("MTU=%d\n", dev_def->devconf.mtu);
	printf("TTL=%d\n", dev_def->devconf.ttl);
	printf("UDPPORT=%d\n", dev_def->devconf.udp_port);
	printf("ONBOOT=yes\n");

	printf("\n");
	switch(dev_def->iface){
	case IF_IFACE_SYNC_SERIAL:
		printf("MEDIA=%s\n",
			INT_DECODE(dev_def->devconf.interface)); 
		if (!dev_def->devconf.bps){
			printf("CLOCK=ext\n");
		}else{
			printf("CLOCK=int\n");
			printf("BAUDRATE=%d", dev_def->devconf.bps);
		}
		break;

	case IF_IFACE_T1:
	case IF_IFACE_E1:
		printf("MEDIA=%s\n", 
			MEDIA_DECODE(dev_def->te_cfg.media));
		printf("LCODE=%s\n", 
			LCODE_DECODE(dev_def->te_cfg.lcode));
		printf("FRAME=%s\n", 
			FRAME_DECODE(dev_def->te_cfg.frame));
		printf("TECLOCK=%s\n",
			TECLK_DECODE(dev_def->te_cfg.te_clock));
		if (dev_def->iface == IF_IFACE_T1){
			printf("LBO=%s\n",
				LBO_DECODE(dev_def->te_cfg.lbo));
		}
		if (dev_def->te_cfg.active_ch == ENABLE_ALL_CHANNELS){
			printf("ACTIVE_CH=all\n");
		}else if (dev_def->te_cfg.active_ch == 0){
			printf("ACTIVE_CH=none\n");
		}else{
			unsigned long active_ch = dev_def->te_cfg.active_ch;
			int first = 1, i;
			int channel_range = 
				(dev_def->iface == IF_IFACE_T1) ?
					NUM_OF_T1_CHANNELS: NUM_OF_E1_CHANNELS;

			printf("ACTIVE_CH=");
			for(i = 0; i < channel_range; i++){
				if (active_ch & (1 << (i - 1))){
					if (!first) printf(".");
					printf("%d", i);
				}
			}
			printf("\n");
		}
                break;

	}	

	printf("\n");
	switch(ifsettings.type){
#if defined(CONFIG_GENERIC_CISCO)
	case IF_PROTO_CISCO:
		printf("PROTO=cisco\n");
#if defined(__LINUX__)
		if (proto.cisco.interval){
			printf("INTERVAL=%d\n",
				proto.cisco.interval);
		}
		if (proto.cisco.timeout > 1){
			printf("TIMEOUT=%d\n",
				proto.cisco.timeout);
		}
#endif
		break;
#endif

#if defined(CONFIG_GENERIC_FR)
	case IF_PROTO_FR:
		printf("PROTO=fr\n");
		switch(proto.fr.lmi){
		case WAN_LMI_NONE:
			printf("LMI=none\n");;
			break;
		case WAN_LMI_ANSI:
		case WAN_LMI_CCITT:
			printf("LMI=%s\n", 
				(proto.fr.lmi == WAN_LMI_ANSI) ? "ansi" : "ccitt");
			printf("T391=%d\n", proto.fr.t391);
			printf("T392=%d\n", proto.fr.t392);
			printf("N391=%d\n", proto.fr.n391);
			printf("N392=%d\n", proto.fr.n392);
			printf("N393=%d\n", proto.fr.n393);
			printf("DCE=%d\n", proto.fr.dce);
		}
		break;
	case IF_PROTO_FR_ADD_PVC:
		break;
	case IF_PROTO_FR_DEL_PVC:
		break;
#endif

#if defined(CONFIG_GENERIC_PPP)
	case IF_PROTO_PPP:
		printf("PROTO=ppp\n");
		break;
#endif
	default:
		return -EINVAL;
	}
	printf("\n");
	return 0;
}

