/****************************************************************************
* wanpipe_sppp_iface.c	WANPIPE(tm) OpenSource WANPIPE Driver.
*
* Author:		Alex Feldman	<al.feldman@sangoma.com>
*		
*
* Copyright:		(c) 2004 Sangoma Technologies Inc.
*
*		This program is free software; you can redistribute it and/or
*		modify it under the terms of the GNU General Public License
*		as published by the Free Software Foundation; either version
*		2 of the License, or (at your option) any later version.
* ============================================================================
* Jan 28, 2004	Alex Feldman	Initial version.
*****************************************************************************/

#if defined (__FreeBSD__) || defined(__OpenBSD__)

#include <net/wanpipe_config.h>
#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/sdlapci.h>
#include <net/wanpipe.h>	/* WANPIPE common user API definitions */
#include <net/if_wanpipe_common.h>
#include <net/wanpipe_generic.h>

#else

# error "Only FreeBSD OS support SPPP interface!"

#endif

/****** Defines & Macros ****************************************************/

#ifdef	_DEBUG_
#define	STATIC
#else
#define	STATIC		static
#endif

/****** Function Prototypes *************************************************/
netdevice_t* wanpipe_generic_register (sdla_t* card);
void wanpipe_generic_unregister (netdevice_t* dev);
static sdla_t* wanpipe_generic_getcard(netdevice_t* dev);
static void wanpipe_generic_init(void* dev);
static int wanpipe_generic_ioctl(netdevice_t* dev, u_long cmd, caddr_t data);
static void wanpipe_generic_watchdog(netdevice_t* dev);
static void wanpipe_generic_start(netdevice_t* dev);
int wanpipe_generic_open(netdevice_t* dev);
int wanpipe_generic_close(netdevice_t* dev);
int wanpipe_generic_input(netdevice_t* dev, netskb_t* skb);

/****** Global Data **********************************************************
 * Note: All data must be explicitly initialized!!!
 */
static char* ifname_format = "hdlc%d";

/******* WAN Device Driver Entry Points *************************************/
static sdla_t* wanpipe_generic_getcard(netdevice_t* dev)
{
	sdla_t*	card;

	if (wan_netif_priv(dev) == NULL){
		DEBUG_EVENT("%s: Invalid device private structure pointer\n",
				wan_netif_name(dev));
		return NULL;
	}
	card = ((sdla_t*)((wanpipe_common_t*)wan_netif_priv(dev))->card);
	if (card == NULL){
		DEBUG_EVENT("%s: Invalid Sangoma device card\n",
				wan_netif_name(dev));
		return NULL;
	}	
	return card;
}

netdevice_t* wanpipe_generic_register (sdla_t* card)
{
	struct sppp*	sppp;
	netdevice_t*	dev;
	static int	ifunit = 0;

	/* Register in HDLC device */
	sppp = (struct sppp*)wan_malloc(sizeof(struct sppp));
	if (sppp == NULL){
		DEBUG_EVENT("%s: Failed allocate SPPP private structure !\n",
				card->devname);
		return NULL;
	}

	dev = (netdevice_t*)sppp;
	sprintf(card->u.c.if_name, ifname_format, ifunit);
	bzero(dev, sizeof(struct sppp));
	wan_netif_init(dev, card->u.c.if_name, ifunit);
	dev->if_softc = NULL;	/* FIXME: Will set from new_if */
	dev->if_mtu = PP_MTU;
	dev->if_flags = IFF_POINTOPOINT | IFF_MULTICAST;
#if defined(__FreeBSD__)
	dev->if_init = wanpipe_generic_init;	/* Will set from new_if() */
#endif
	dev->if_ioctl = wanpipe_generic_ioctl;	/* Will set from new_if() */
	dev->if_start = wanpipe_generic_start;
	dev->if_watchdog = wanpipe_generic_watchdog;
	/*dev->if_snd.ifq_maxlen = ifqmaxlen;*/

	IFATTACH(dev);
	sppp_attach(dev);
	BPF_ATTACH(dev, DLT_PPP);

	DEBUG_EVENT("%s: New interface %s allocated!\n", 
					card->devname, wan_netif_name(dev));
	ifunit ++;
	return dev;
}

void wanpipe_generic_unregister (netdevice_t* dev)
{
	sdla_t*	card;

	if ((card = wanpipe_generic_getcard(dev)) == NULL){
		return;
	}
	DEBUG_EVENT("%s:%s: Unregister interface!\n", 
				card->devname, wan_netif_name(dev));
	sppp_detach(dev);
	IFDETACH(dev);
	wan_free((struct sppp*)dev);
}


static void wanpipe_generic_start(netdevice_t* dev)
{
	sdla_t*		card;
	netskb_t*	opkt;
	int		s = splimp(), err = 0;

	DEBUG_TEST("%s: Sending new packet...\n",
			wan_netif_name(dev));
	if ((card = wanpipe_generic_getcard(dev)) == NULL){
		return;
	}
	while(1){
		if (WAN_SPPP_ISEMPTY(dev)){
			/* No more packets in send queue */
			break;
		}
						
		if ((opkt = WAN_SPPP_DEQUEUE(dev)) == NULL){
			/* Should never happened, packet pointer is NULL */
			break;
		}
		if (card->wandev.protocol_send == NULL){
			DEBUG_EVENT("%s: Protocol send function is undefined!\n",
					wan_netif_name(dev));
			break;
		}
		err = card->wandev.protocol_send(opkt, dev);
		if (err){
			IF_DROP (&dev->if_snd);
			wan_skb_free(opkt);	/*m_freem(opkt);*/
			break;
		}
		/* report the packet to BPF if present and attached */
		wan_bpf_report(dev, opkt, 1);
	}
	return;
}


static int wanpipe_generic_ioctl(netdevice_t* dev, u_long cmd, caddr_t data)
{
	sdla_t*			card = NULL;
	wanpipe_common_t*	common;
	struct if_settings	ifsettings;
	short			if_flags;
	int			err = 0, s;
	int			was_up, should_be_up;

	if ((card = wanpipe_generic_getcard(dev)) == NULL){
		return -EINVAL;
	}
	s = splimp();
	common = wan_netif_priv(dev);
	DEBUG_TEST("%s: Wanpipe generic ioctl \n",
				wan_netif_name(dev));
	if_flags = wan_netif_flags(dev);
	was_up = if_flags & IFF_UP;
	should_be_up = (if_flags & IFF_UP) || (cmd == SIOCSIFADDR);
	if (!was_up && should_be_up){
		/* Interface should be up -- start it. */
		wan_netif_set_flags(dev, (if_flags | IFF_UP));
		wanpipe_generic_start(dev);
	}else if(was_up && !should_be_up){
		/* Interface should be down -- stop it. */
		if_flags &= ~IFF_UP;
		wan_netif_set_flags(dev, if_flags);
	}
	switch(cmd){
	case SIOCSIFADDR:
		wanpipe_generic_init(dev); 
		err = 1;
		break;

	case SIOCSIFFLAGS:
		DEBUG_IOCTL("%s:%s: ioctl(SIOCSIFFLAGS) called.\n", 
				card->devname, wan_netif_name(dev));
	    	/*
     		** If the interface is marked up - enable communications. 
	     	** If down - disable communications.  IFF_UP is taken 
		** care of before entering this function.
	     	*/
    		if ((dev->if_flags & IFF_UP) == 0){ 
			/* bring it down */
			DEBUG_EVENT("%s:%s: Bringing interface down.\n", 
					card->devname, wan_netif_name(dev));
			dev->if_flags &= ~IFF_RUNNING;
		}else{ /* bring it up */ 
			struct sppp* sp = (struct sppp*)dev;
			DEBUG_EVENT("%s:%s: Bringing interface up.\n", 
					card->devname, wan_netif_name(dev));
			if (sp->pp_up){
				sp->pp_up(sp);
			}
		}
		break;

	case SIOC_WANPIPE_DEVICE:
		err = WAN_COPY_FROM_USER(&ifsettings, 
					((struct ifreq*)data)->ifr_data,
					 sizeof(struct if_settings));
		if (err){
			DEBUG_EVENT("%s: Failed to copy from user space!\n",
						card->devname);
			goto wanpipe_generic_ioctl_done;
		}
		if (ifsettings.type == IF_GET_PROTO){
			ifsettings.type = common->proto;
			err = WAN_COPY_TO_USER(((struct ifreq*)data)->ifr_data,
						&ifsettings,
						sizeof(struct if_settings));
			if (err){
				DEBUG_EVENT("%s: Failed to copy to user space!\n",
							card->devname);
			}
			goto wanpipe_generic_ioctl_done;
			
		}
		((struct ifreq*)data)->ifr_data = (caddr_t)&ifsettings;
		break;
	}
	if (card->wandev.protocol_ioctl){
		/* Argument seqeunce is change for Linux order */
		err = card->wandev.protocol_ioctl(dev, (struct ifreq*)data, cmd);
	}
	switch(cmd){
	case SIOC_WANPIPE_DEVICE:
		switch (ifsettings.type) {
		case IF_PROTO_CISCO:
			common->proto = IF_PROTO_CISCO;
#if defined(__FreeBSD__)
			if_flags = wan_netif_flags(dev);
			if_flags |= IFF_LINK2;
			wan_netif_set_flags(dev, if_flags);
#elif defined(__OpenBSD__)
			((struct sppp *)dev)->pp_flags |= PP_CISCO;
#endif
			((struct sppp *)dev)->pp_flags |= PP_KEEPALIVE;
			break;
		case IF_PROTO_PPP:
			common->proto = IF_PROTO_PPP;
#if defined(__FreeBSD__)
			if_flags = wan_netif_flags(dev);
			if_flags &= ~IFF_LINK2;
			wan_netif_set_flags(dev, if_flags);
#elif defined(__OpenBSD__)
			((struct sppp *)dev)->pp_flags &= ~PP_CISCO;
#endif
			((struct sppp *)dev)->pp_flags &= ~PP_KEEPALIVE;
			break;
		}
		break;
	}
	if (err == 1){
		err = WAN_SPPP_IOCTL(dev, 
			(cmd==SIOC_WANPIPE_DEVICE) ? SIOCSIFFLAGS : cmd,
			data);
	}
wanpipe_generic_ioctl_done:
	splx(s);
	return err;
}

static void wanpipe_generic_init(void* dev)
{
	sdla_t*	card;

	if ((card = wanpipe_generic_getcard(dev)) == NULL){
		return;
	}
	if (card->wandev.protocol_open){
		card->wandev.protocol_open((netdevice_t*)dev);
	}
	return;
}

static void wanpipe_generic_watchdog(netdevice_t* dev)
{
	return;
}

int wanpipe_generic_open(netdevice_t* dev)
{
	return 0;
}

int wanpipe_generic_close(netdevice_t* dev)
{
	return 0;
}

int wanpipe_generic_input(netdevice_t* dev, netskb_t* skb)
{
	wan_bpf_report(dev, skb, 1);
#if defined(NETGRAPH)
	if (card->hook){
		int error;
		NG_SEND_DATA_ONLY(error, 
				  card->hook, 
				  skb);
		card->inbytes += wan_skb_len(len);
	}
#else
	sppp_input(dev, skb);
#endif
	return 0;
}

/************************************ END **********************************/
